
//
// StreamT.cc
//
//
// HISTORY:
//
//    Date        Author   Modifications
//    --------    ------   -------------
//    15/06/97    MT       Initial Version
//
//
//
//
//
//
//
// DESCRIPTION
//
//
//




#include "StreamT.h"
#include "Tracer.h"


StreamT::StreamT(
		istream &	thisStream,
		const string &	fileName,
		unsigned int	lineNumber,
		unsigned int	columnNumber ) :

		_is( thisStream ),
		_fileName( fileName ),
		_currentLineNumber(lineNumber),
		_currentColumnNumber(columnNumber)
{
	PROC
	ENDPROC
}


char
StreamT::GetNextChar()
{

	PROC

	char	thisChar;


	thisChar = _is.get();


	if( thisChar == '\n' )
	{
		_currentLineNumber++;
		_currentColumnNumber = 1;

	}
	else
	{
		_currentColumnNumber++;

	} /* if */

	RETURN thisChar;

}


char
StreamT::peek()
{
	PROC
	RETURN _is.peek();
}


void
StreamT::getLine( char * line, streamsize n )
{
	PROC
	_is.getline( line, n );
	_currentLineNumber++;
	_currentColumnNumber = 1;
	ENDPROC
}


void
StreamT::skipLine()
{
	PROC
	char	line[ 1024 ];
	_is.getline( line, 1024 );
	_currentLineNumber++;
	_currentColumnNumber = 1;
	ENDPROC
}


char
StreamT::GetNextEscapedChar()
{
	PROC

	char	thisChar;

	thisChar = _is.get();

	switch( thisChar )
	{
		case 'n' :

			thisChar = '\n'; break;

		case 't' :

			thisChar = '\t'; break;

		case 'v' :

			thisChar = '\v'; break;

		case 'b' :

			thisChar = '\b'; break;

		case 'r' :

			thisChar = '\r'; break;

		case 'f' :

			thisChar = '\f'; break;

		case 'a' :

			thisChar = '\a'; break;

		case '\\' :

			thisChar = '\\'; break;

		case '?' :

			thisChar = '?'; break;

		case '"' :

			thisChar = '"'; break;

		case '>' :

			thisChar = '>'; break;

		case '\'' :

			thisChar = '\''; break;

		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
		case 7:
		case 8:
		case 9:
			// ******************
			// To be Implemented.
			// ******************
			break;

		case 'x':

			break;

		default :

			break;

	} /* switch */

	RETURN thisChar;

}


void
StreamT::UngetCharacter( char		thisCharacter )
{
	PROC

	//	The following code is probably redundant.

	if( thisCharacter == EOF)
	{
		ALTRETURN;
	} /* if */

       if( !_is.unget() )
       {
		cerr << _fileName << ": Unget character failed: "
		<< " Line = " << _currentLineNumber
		<<  " Column = " << _currentColumnNumber
		<< " Character = '" << thisCharacter << "'"
		<< " Character Value = " << int( thisCharacter )
		<< endl;
		cerr << "Stream = \"";
		cerr << *this;
		exit( 1 );
	} // if 


	if( thisCharacter == '\n')
	{
		_currentLineNumber--;
		_currentColumnNumber = 1; // This leads to a bug.

	}
	else
	{
		_currentColumnNumber--;

	} /* if */

	ENDPROC
}


int
StreamT::EndOfFile() const
{
	PROC
	RETURN _is.eof();

}


unsigned int
StreamT::CurrentLineNo() const
{
	PROC
	RETURN _currentLineNumber;

}


unsigned int
StreamT::CurrentColumnNo() const
{
	PROC
	RETURN _currentColumnNumber;

}


ostream &
operator<<( ostream & os, const StreamT & s )
{
	PROC
	s._is.clear();
	const unsigned int oldpos = s._is.tellg();

	for( unsigned int i = 0; s._is.seekg( i ) && !s._is.eof(); ++i ) os << char(s._is.peek());
	s._is.clear();
	s._is.seekg( oldpos );
	RETURN os;
}
