/* Here we create an extended version of DataType from Dal.h which
   adds signed and unsigned long integers. These are not valid
   container types for the dal, but we need to add them so that
   interpr can define unsigned long arrays which are used to define
   dimensions in certain dal calls, such as addColumn. The test
   harnesses defined the dimensions arrays using Int32 pointers which
   worked on 32-bit systems, but this causes run-time problems in
   64-bit. I'm not sure why the compiler doesn't warn about type
   mismatches - it may be that we are hiding things in void * so it
   has no way of knowing.

   14-MAR-2013 Initial Version (EC) */

#ifndef DAL_TYPES_EXTENDED_H
#define DAL_TYPES_EXTENDED_H

enum interprDataType {
  Bool = 1, /**< Boolean */
  Int8, /**< 8-bit unsigned integers */
  Uint16, /**< 16-bit unsigned integers */
  Int16, /**< 16-bit signed integers */
  Uint32, /**< 32-bit unsigned integers */
  Int32, /**< 32-bit signed integers */
  Real32, /**< 32-bit reals */
  Real64, /**< 64-bit reals */
  DString, /**< Strings of characters */

  // Here are the new types, which may or may not be identical to the
  // int32 types, depending on the architecture

  Ulong,
  Long
};

#endif
