//
// scanner.cc
//
//
// HISTORY:
//
//    Date        Author   Modifications
//    --------    ------   -------------
//    15/06/97    MT       Initial Version
//
//
//
//
//
//
//
// DESCRIPTION
// Lexical analyser.
//
//
//


#include <cstring>
#include <cctype>
#include <cstdlib>
#include <cstdarg>
#include <iostream>
#include <sstream>

#include "Tracer.h"
#include "errstr.h"
#include "token.h"
#include "file.h"
#include "scanner.h"



ScannerT::ScannerT(
		istream & 	inputStream,
		char		commentDesignator,
		const string &	fileName,
		unsigned int	lineNumber,
		unsigned int	columnNumber)

	:	File(*( new StreamT( inputStream, fileName, lineNumber, columnNumber ))),
		State( NORMAL ),
		_commentDesignator( commentDesignator )
{

	PROC

	FetchNextToken();

	ENDPROC

}


ScannerT::ScannerT(
		const string &	fileName,
		char		commentDesignator)

	:	File(*( new FileT( fileName ))),
		State( NORMAL ),
		_commentDesignator( commentDesignator )
{

	PROC

	FetchNextToken();

	ENDPROC

}


ScannerT::~ScannerT()
{
	PROC

	delete &File;

	ENDPROC

}



Token
ScannerT::acceptNext()
{
	PROC

	const Token nextToken = LookaheadToken;

	FetchNextToken();

	RETURN nextToken;

}


void
ScannerT::getLine( char * line, streamsize n )
{
	PROC

	char	temp[ n ];

	File.getLine( temp, n );
	strcpy( line, LookaheadToken.rawValue().c_str() );
	strcat( line, temp );
	FetchNextToken();

	ENDPROC

}


Token
ScannerT::Next() const
{
	PROC
	RETURN LookaheadToken;

}


bool
ScannerT::NextIs( char	ThisCharacter ) const
{
	PROC

	if( Token::Character == LookaheadToken.id() )
	{
		if( ThisCharacter == LookaheadToken.asCharacter() )
		{
			RETURN true;

		} // if
	} // if

	RETURN false;

}


bool
ScannerT::NextIs( char *	ThisString ) const
{
	PROC

	if( Token::Identifier == LookaheadToken.id() )
	{
		if( LookaheadToken.asString() == ThisString )
		{
			RETURN true;

		} // if
	} // if

	RETURN false;

}


bool
ScannerT::nextIs( const string & thisString ) const
{
	PROC

	if( Token::Identifier == LookaheadToken.id() )
	{
		if( LookaheadToken.asString() == thisString )
		{
			RETURN true;

		} // if
	} // if

	RETURN false;

}


bool
ScannerT::nextIs( Token::IDType		tokenID ) const
{
	PROC

	if( tokenID == LookaheadToken.id() )
	{
		RETURN true;
	} // if

	RETURN false;

}


void
ScannerT::Match(
	char *			String,
	Token &			ThisToken )
{
	PROC

	if( LookaheadToken.id() == Token::Identifier )
	{
		if(
			LookaheadToken.asString() == String )
		{
			ThisToken = LookaheadToken;

			FetchNextToken();

			ALTRETURN;

		} // if
	} // if

	cerr <<
		File.FileNameOf() <<
		": Error at line " <<
		File.CurrentLineNo() <<
 		", and at or near column " <<
		File.CurrentColumnNo() <<
		" - Expected string " <<
		String <<
		endl;

	exit( 1 );

}


void
ScannerT::match(
	const string &		thisIdentifier )
{
	PROC

	if( LookaheadToken.id() == Token::Identifier )
	{
		if(
			LookaheadToken.asString() == thisIdentifier )
		{
			FetchNextToken();

			ALTRETURN;

		} // if
	} // if

	cerr <<
		File.FileNameOf() <<
		": Error at line " <<
		File.CurrentLineNo() <<
 		", and at or near column " <<
		File.CurrentColumnNo() <<
		" - Expected string " <<
		thisIdentifier <<
		endl;

	exit( 1 );

}


void
ScannerT::Match(
	Token::IDType		ThisTokenID,
	Token &			ThisToken )
{
	PROC

	if ( ThisTokenID == LookaheadToken.id() )
	{
		ThisToken = LookaheadToken;

		FetchNextToken();
	}
	else
	{
		cerr <<
			File.FileNameOf() <<
			": Error at line " <<
			File.CurrentLineNo() <<
 			", and at or near column " <<
			File.CurrentColumnNo() <<
			" - Expected tokenID " <<
			int( ThisTokenID ) <<
			endl;

		exit( 1 );

	} // if

	ENDPROC
}


void
ScannerT::Match(
	char		ThisChar,
	Token &	ThisToken )
{
	PROC

	if( Token::Character == LookaheadToken.id() )
	{
		if( ThisChar == LookaheadToken.asCharacter() )
		{
			ThisToken = LookaheadToken;

			FetchNextToken();
		}
		else
		{
			cerr <<
				File.FileNameOf() <<
				": Error at line " <<
				File.CurrentLineNo() <<
 				", and at or near column " <<
				File.CurrentColumnNo() <<
				" - Expected character " <<
				ThisChar <<
				endl;

			exit( 1 );

		} // if
	}
	else
	{
		cerr <<
			File.FileNameOf() <<
			": Error at line " <<
			File.CurrentLineNo() <<
 			", and at or near column " <<
			File.CurrentColumnNo() <<
			" - Expected character " <<
			ThisChar <<
			endl;

		exit( 1 );

	} // if

	ENDPROC

}


void
ScannerT::match( char	thisChar )
{
	PROC

	if( Token::Character == LookaheadToken.id() )
	{
		if( thisChar == LookaheadToken.asCharacter() )
		{
			FetchNextToken();
		}
		else
		{
			cerr <<
				File.FileNameOf() <<
				": Error at line " <<
				File.CurrentLineNo() <<
 				", and at or near column " <<
				File.CurrentColumnNo() <<
				" - Expected character " <<
				thisChar <<
				endl;

			exit( 1 );

		} // if
	}
	else
	{
		cerr <<
			File.FileNameOf() <<
			": Error at line " <<
			File.CurrentLineNo() <<
 			", and at or near column " <<
			File.CurrentColumnNo() <<
			" - Expected character " <<
			thisChar <<
			endl;

		exit( 1 );

	} // if

	ENDPROC

}


Token
ScannerT::matchOneOf(
	Token::IDType	tokenID,
	... )
{
	PROC

	va_list		ptrToValue;
	bool		matched = false;
	Token		token;
	Token::IDType	next;


	if( nextIs( tokenID ))
	{
		Match( tokenID, token );
		matched = true;
	}
	else
	{
		va_start( ptrToValue, tokenID );
		next = (Token::IDType)va_arg( ptrToValue, int);

		while( next != Token::Undefined )
		{
			if( nextIs( next ))
			{
				Match( next, token );
				matched = true;
				break;
			} // if

			next = (Token::IDType)va_arg( ptrToValue, int);
		} // while

		va_end( ptrToValue );

	} // if


	if( !matched )
	{

		cerr << File.FileNameOf()
			<< ": Error at line "
			<< File.CurrentLineNo()
			<<  ", and at or near column "
			<< File.CurrentColumnNo()
			<< "."
			<< endl;

		cerr << "Expected one of the following: "
			<< endl;

		cerr << int( tokenID ) << " ";

		va_start( ptrToValue, tokenID );
		next = (Token::IDType)va_arg( ptrToValue, int);

		while( next != Token::Undefined )
		{
			cerr << int( next ) << " ";
			next = (Token::IDType)va_arg( ptrToValue, int);

		} // while

		cerr << endl;

		va_end( ptrToValue );

		exit( 1 );

	} // if


	RETURN token;

}


Token
ScannerT::MatchOneOf(
	char *				String,
	... )
{
	PROC

	va_list		PtrToValue;
	bool		Matched = false;
	Token		Token;
	char *		NextString;



	if( NextIs( String ))
	{
		Match( String, Token );
		Matched = true;
	}
	else
	{
		va_start( PtrToValue, String );

		NextString = va_arg( PtrToValue, char * );

		while( NextString != 0 )
		{
			if( NextIs( NextString ))
			{
				Match( NextString, Token );
				Matched = true;
				break;
			} // if

			NextString = va_arg( PtrToValue, char * );
		} // while

		va_end( PtrToValue );

	} // if


	if( !Matched )
	{

		cerr << File.FileNameOf()
			<< ": Error at line "
			<< File.CurrentLineNo()
			<<  ", and at or near column "
			<< File.CurrentColumnNo()
			<< "."
			<< endl;

		cerr << "Expected one of the following: "
			<< endl;

		cerr << String << " ";

		va_start( PtrToValue, String );
		NextString = va_arg( PtrToValue, char * );

		while( NextString != 0 )
		{
			cerr << NextString << " ";
			NextString = va_arg( PtrToValue, char * );

		} // while

		cerr << endl;

		va_end( PtrToValue );

		exit( 1 );

	} // if


	RETURN Token;

}


void
ScannerT::matchSignedNumeric(
		Token::IDType		tokenID,
		Token&			token )
{
	PROC
	bool negative = false;

	if( NextIs( '-' ) )
	{
		negative = true;
		match( '-' );
	}
	else if( NextIs( '+' ) )
	{
		match( '+' );
	} // if

	Token numeric;
	Match( tokenID, numeric );

	if( negative )
	{
		switch( numeric.id() )
		{
			case Token::Integer :
				numeric = -numeric.asInt();
				break;
			case Token::Real :
				numeric = -numeric.asReal();
				break;
			default :
				errstr << "Invalid switch" << Sas::error( "internalError" );
				break;
		} // switch
	} // if
	token = numeric;
	ENDPROC
}


void
ScannerT::SkipCommentsAndWhiteSpace()
{

	PROC

	char		currentChar;

	while( 1 )
	{
		while( isspace( currentChar = File.GetNextChar() )) ;

		// Skip comments (if any)
		if( currentChar == _commentDesignator )
		{
			File.skipLine();
		}
		else
		{
			File.UngetCharacter( currentChar );

			break;

		} // if
	} // while

	ENDPROC

}


void
ScannerT::FetchNextToken()
{

	PROC

	// *******************************
	// Check the state of the scanner.
	// *******************************
	if( State == ScannerT::LINE )
	{
		_lineScanner -> FetchNextToken();
		LookaheadToken = _lineScanner -> LookaheadToken;

		ALTRETURN;

	} // if

	char		thisChar;
	int		charPos = 0;
	char		thisString[ 255 ];


	if( State == FREE_FORMAT_TEXT )
	{
		FetchFreeText();

		ALTRETURN;

	} // if


	SkipCommentsAndWhiteSpace();

	// **********************
	// Check for end of file.
	// **********************
	if( File.EndOfFile() )
	{
		LookaheadToken = Token( Token::EndOfFile );

		ALTRETURN;

	} // if


	// ************************************
	// Set the Token raw value information.
	// ************************************

	LookaheadToken.setDebug( File.FileNameOf() );


	thisChar = File.GetNextChar();

	// ***********************************
	// Check for literal string.
	// A literal string is delimited by ".
	// ***********************************
	if( thisChar == '"' )
	{
		thisChar = File.GetNextChar();

		while( thisChar != '"' )
		{
			if( thisChar == '\\' )
			{
				thisChar = File.GetNextEscapedChar();

			} /* if */

			thisString[ charPos++ ] = thisChar;
			thisChar = File.GetNextChar();

		} // while

		thisString[ charPos ] = '\0';

		LookaheadToken = string( thisString );

		ALTRETURN;

	}
	// ****************************
	// Check for literal character.
	// ****************************
	else if( thisChar == '\'' )
	{
		thisChar = File.GetNextChar();

		if( thisChar == '\\' )
		{
			thisChar = File.GetNextEscapedChar();
		} /* if */


		LookaheadToken = Token( thisChar, Token::QuotedCharacter );

		// =======================
		// Should now check for '.
		// =======================
		thisChar = File.GetNextChar();

		ALTRETURN;

	}
	/*************************/
	/* Check for identifier. */
	/*************************/
	else if( isalpha( thisChar ))
	{
		while( isalnum( thisChar ) || thisChar == '_' )
		{
			thisString[ charPos++ ] = thisChar;
			thisChar = File.GetNextChar();
		} // while

		File.UngetCharacter( thisChar );
		thisString[ charPos ] = '\0';

		LookaheadToken = Token( string( thisString ), Token::Identifier );

		ALTRETURN;

	}
	// *****************************************************
	// Check for numeric: This may be an integer or a float,
	// A real is designated with the decimal point '.'
	// and/or an optional exponent is possible for reals, which
	// is designated by 'e' or 'E'; the exponent itself
	// consists of an optional sign ('+' or '-') followed by
	// an integer (at least one digit is expected).
	// *****************************************************
	else if(
		isdigit( thisChar )
		||
		thisChar == '.' )
	{
		bool		realNumber = false;

		// *****************
		// Get integer part.
		// *****************
		while( isdigit( thisChar ) )
		{
			thisString[ charPos++ ] = thisChar;
			thisChar = File.GetNextChar();

		} // while

		// ********************************
		// Look for optional decimal point.
		// ********************************
		if( thisChar == '.' )
		{
			realNumber = true;
			thisString[ charPos++ ] = thisChar;

			thisChar = File.GetNextChar();

			while( isdigit( thisChar ) )
			{
				thisString[ charPos++ ] = thisChar;
				thisChar = File.GetNextChar();

			} // while
		} // if

		// *************************
		// Now check for optional e.
		// *************************
		if( thisChar == 'e' || thisChar == 'E' )
		{
			realNumber = true;
			thisString[ charPos++ ] = thisChar;
			thisChar = File.GetNextChar();
			// *********************************
			// Now check for optional '+' or '-'
			// *********************************
			if( thisChar == '+' || thisChar == '-' )
			{
				thisString[ charPos++ ] = thisChar;
				thisChar = File.GetNextChar();
			} // if

			// **********************************
			// Must be at least one digit forming
			// exponent value.
			// **********************************
			if( isdigit( thisChar ) )
			{
				while( isdigit( thisChar ) )
				{
					thisString[ charPos++ ] = thisChar;
					thisChar = File.GetNextChar();

				} // while
			}
			else
			{
				cerr << File.FileNameOf() <<
					": Error at line " <<
					File.CurrentLineNo() <<
 					", and at or near column " <<
					File.CurrentColumnNo() <<
					" - Expected at least one digit." <<
					endl;

				exit( 1 );

			} // if
		} // if

		File.UngetCharacter( thisChar );
		thisString[ charPos ] = '\0';

		if( realNumber )
		{
			LookaheadToken = atof( thisString );

		}
		else
		{
			LookaheadToken = atoi( thisString );

		} // if

		ALTRETURN;

	}
	else if( thisChar == '+' )
	{
		char nextChar = File.GetNextChar();

		if( nextChar == '+' )
		{
			LookaheadToken = Token::PlusPlus;
		}
		else
		{
			File.UngetCharacter( nextChar );
			LookaheadToken = thisChar;
		}// if

		ALTRETURN;

	}
	else if( thisChar == '-' )
	{
		char nextChar = File.GetNextChar();

		if( nextChar == '-' )
		{
			LookaheadToken = Token::MinusMinus;
		}
		else
		{
			File.UngetCharacter( nextChar );
			LookaheadToken = thisChar;
		}// if

		ALTRETURN;

	}
	else if( thisChar == '<' )
	{
		char nextChar = File.GetNextChar();

		if( nextChar == '=' )
		{
			LookaheadToken = Token::LessEqual;
		}
		else
		{
			File.UngetCharacter( nextChar );
			LookaheadToken = thisChar;
		}// if

		ALTRETURN;

	}
	else if( thisChar == '>' )
	{
		char nextChar = File.GetNextChar();

		if( nextChar == '=' )
		{
			LookaheadToken = Token::GreaterEqual;
		}
		else
		{
			File.UngetCharacter( nextChar );
			LookaheadToken = thisChar;
		}// if

		ALTRETURN;

	}
	else if( thisChar == '=' )
	{
		char nextChar = File.GetNextChar();

		if( nextChar == '=' )
		{
			LookaheadToken = Token::EqualEqual;
		}
		else
		{
			File.UngetCharacter( nextChar );
			LookaheadToken = thisChar;
		}// if

		ALTRETURN;

	}
	else if( thisChar == '!' )
	{
		char nextChar = File.GetNextChar();

		if( nextChar == '=' )
		{
			LookaheadToken = Token::NotEqual;
		}
		else
		{
			File.UngetCharacter( nextChar );
			LookaheadToken = thisChar;
		}// if

		ALTRETURN;

	}
	else
	{
		LookaheadToken = thisChar;

		ALTRETURN;

	} // if

	// **********************
	// Check for end of file.
	// **********************
	if( File.EndOfFile() )
	{
		LookaheadToken = Token( Token::EndOfFile );

	} // if

	ENDPROC

}


void
ScannerT::FetchFreeText()
{

	PROC

	int			charPos = 0;
	char			thisChar;
	char			thisString[ 8192 ];


	while( (thisChar = File.GetNextChar()) != '>' )
	{
		// Ignore formatting characters
		if(
			thisChar == '\n' ||
			thisChar == '\t' ||
			thisChar == '\r' ) continue;

		if( thisChar == '\\' )
		{
			thisChar = File.GetNextEscapedChar();
		} // if

		thisString[ charPos++ ] = thisChar;

	} // while

	thisString[ charPos ] = '\0';

	File.UngetCharacter( thisChar );

	LookaheadToken = Token( string( thisString ), Token::FreeText );


	ENDPROC

}


bool
ScannerT::EndOfFile() const
{
	PROC
	RETURN LookaheadToken.id() == Token::EndOfFile;

}


void
ScannerT::SetState(
		StateType		toThisState )
{
	PROC
	switch( toThisState )
	{
		case ScannerT::NORMAL :
		{
			if( State == LINE )
			{
				State = toThisState;
				//delete _line;
				//delete _lineStr;
				//delete _lineScanner;
				FetchNextToken();
			} // if
		}
			break;
		case ScannerT::FREE_FORMAT_TEXT :
			break;
		case ScannerT::LINE :
		{
			// ************************************************************
			// Initialise a scanner to scan the line, in such a way that
			// the filename, current column and line numbers are preserved.
			// ************************************************************
			char * 			_line = new char[ 1024 ];
			const unsigned int	currentLineNumber = stream().CurrentLineNo();
			const unsigned int	currentColumnNumber = stream().CurrentColumnNo();
			char	temp[ 1024 ];

			File.getLine( temp, 1024 );
			strcpy( _line, LookaheadToken.rawValue().c_str() );
			strcat( _line, temp );
			istringstream *	 _lineStr = new istringstream( _line);
			_lineScanner = new ScannerT(
						*_lineStr,
						commentDesignator(),
						File.FileNameOf(),
						currentLineNumber,
						currentColumnNumber );
			LookaheadToken = _lineScanner -> LookaheadToken;
		}
			break;
		default :
			cerr << "Unrecognised Scanner State Type: " <<
			int( toThisState ) << endl;
			exit( 1 );
	} // switch

	State = toThisState;

	ENDPROC
}


ScannerT::StateType
ScannerT::StateOf() const
{
	PROC
	RETURN State;
}



